/****************************************************************** 
* Copyright (c) 2003, Mario Hochreiter, Student of 
* Software Engineering at FH-Hagenberg, Austria. 
* Project member of the isac team at the Institute for 
* Software-Technologie, Graz University of Technology, Austria. 
* 
* Use is subject to license terms.
******************************************************************/

package isac.gui.util;

import isac.util.formulae.CalcHead;
import isac.util.formulae.CalcFormula;
import isac.util.formulae.Position;
import isac.util.interfaces.ICalcElement;
import isac.util.interfaces.ICalcIterator;

import java.io.Serializable;
import java.rmi.RemoteException;
import java.util.Iterator;
import java.util.Vector;
import org.apache.log4j.Logger;

/**
 * CalcModelHierarchy is the "glue" between the <code>CalculationModel</code>
 * and the <code>CalcIterator</code>. This class holds
 * Objects from type <code>CalcModelNode</code> and is
 * responsible to build up a hierarchy that represents the  
 * <code>CalcElement</code>'s in the <code>CalcTree</code>.
 *  
 * @author Mario Hochreiter
 * @version 1.0
 * 
 * @author Alois Kirchsteiger
 * @version 1.1
 */

public class CalcModelHierarchy implements Serializable {
	static Logger logger = Logger.getLogger(CalcModelHierarchy.class.getName());
  
  CalcModelNode root_;
  /**
   * @author Alois Kirchsteiger
   * creates the hierachy
   * the nodes are stored in an ArrayList together with their levels
   */
  public CalcModelHierarchy() {
    root_ = new CalcModelNode(null); //root of the worksheet, not visible for the user
  }
  
  /**
   * @return the root of the worksheet
   */
  public CalcModelNode getRoot() {
    return root_;
  }
  
  /**
   * @author Alois Kirchsteiger
   * @return Object[3] for treeEvents
   * the result has the following structure:
   * result[0] = path to the parent
   * result[1] = int[1] with the index of the inserted Node as only element
   * result[2] = Object[1] with the inserted Node as only element
   * 
   * returns null if unsuccessful, e.g. because the node is already in the tree
   */
  public Object[] insert(ICalcIterator it) {
    if (getNodeForValue(it, root_) != null) {
      return null;
    }
    CalcModelNode parent = getParent(it, root_);
    CalcModelNode new_node = new CalcModelNode(it);
    parent.add(new_node);
    return new Object[] { parent.getPath().toArray(), 
                          new int[]{new_node.getIndex()},
                          new Object[]{new_node} };
  }
  
  /**
   * @author Alois Kirchsteiger
   * @return Object[3] for treeEvents
   * the result has the following structure:
   * result[0] = path to the parent
   * result[1] = int[1] with the index of the inserted Node as only element
   * result[2] = Object[1] with the inserted Node as only element
   * 
   * returns null if unsuccessful, e.g. because the node does not exist
   */
   public Object[] delete(ICalcIterator it) {
     CalcModelNode node_to_delete = getNodeForValue(it, root_);
     if (node_to_delete == null) {
       return null;
     }
     CalcModelNode parent = node_to_delete.getParent();
     int index = node_to_delete.getIndex();
     parent.remove(node_to_delete);
     return new Object[] {parent.getPath().toArray(),
                          new int[] {index}, 
                          new Object[] {node_to_delete}};
   }
  
  /**
   * @param it
   * @return the parent node under which the new node with this iterator will be inserted
   */
  private CalcModelNode getParent(ICalcIterator it, CalcModelNode node) {
    try {
      if ((node.getValue() == null && it.getLevel() == 0) ||
          (node.getValue() != null && node.getValue().getLevel()  == (it.getLevel()-1))) {
        if (node.getValue() == null || compareIterators(node.getValue(), it) < 0) {
          return node;
        } else {
          return null;
        }
      }
      Iterator iterator = node.getChildren().iterator();
      CalcModelNode current_best_node = null;
      while (iterator.hasNext()) {
        CalcModelNode current_node = (CalcModelNode)iterator.next();
        CalcModelNode temp = getParent(it, current_node);
        if (temp != null) {
          current_best_node = temp;
        } else {
          return current_best_node;
        }
      }
      return current_best_node;
    } catch (RemoteException e) {
      e.printStackTrace();
    }
    return node;
  }
  
  /**
   * @param it
   * @return the path to the father of this node
   */
  public Object[] getParentPath(ICalcIterator it) {
    CalcModelNode node = getNodeForValue(it, root_);
    Vector path = node.getParentPath();
    if (path.size() != 0) {
      return path.toArray();
    } else {
      return null;
    }
  }
  
  /**
   * @param it
   * @return the path to the node containing it
   */
  public Object[] getPath(ICalcIterator it) {
    CalcModelNode node = getNodeForValue(it, root_);
    Vector path = node.getPath();
    if (path.size() != 0) {
      return path.toArray();
    } else {
      return null;
    }
  }
  
  /**
   * @param it
   * @return the index of the node that contains the iterator
   */
  public int getIndex(ICalcIterator it) {
    CalcModelNode node = getNodeForValue(it, root_);
    return node.getIndex();
  }
  
  /**
   * @param node
   * @return the number of childs for the node
   */
  public int getChildCount(CalcModelNode node) {
    return node.getChildCount();
  }
  
  /**
   * @param it
   * @return the node which has the specified value, null if node not found
   */
  private CalcModelNode getNodeForValue(ICalcIterator it, CalcModelNode node) {
    if (node.getValue() != null && compareIterators(node.getValue(), it) == 0) {
      return node;
    }
    CalcModelNode result = null;
    Iterator childIterator = node.getChildren().iterator();
    while (childIterator.hasNext()) {
      CalcModelNode currentNode = (CalcModelNode) childIterator.next();
      result = getNodeForValue(it, currentNode);
      if (result != null) {
        return result;
      }
    }
    return null;
  }
  
   //FIXME: LK remove as soon as iterator can compare positions
  public static int compareIterators(ICalcIterator it1, ICalcIterator it2){
    try {
    ICalcElement f1 = it1.getFormula();
    ICalcElement f2 = it2.getFormula();
    Position p1 = null;
    Position p2 = null;
    if (f1 instanceof CalcFormula) {
      p1 = ((CalcFormula)f1).getPosition();
    } else {
      p1 = ((CalcHead)f1).getPosition();
    }
    if (f2 instanceof CalcFormula) {
      p2 = ((CalcFormula)f2).getPosition();
    } else {
      p2 = ((CalcHead)f2).getPosition();
    }
    return p1.compareTo(p2);
    } catch (RemoteException e) {
      return 0;
    }
  }

  /**
   * resets the tree
   */
  public void reset() {
    root_ = null;
    root_ = new CalcModelNode(null);
  }

  /**
   * @author Alois Kirchsteiger
   * @return Object[3] for treeEvents with an empty node
   * the result has the following structure:
   * result[0] = path to the parent
   * result[1] = int[1] with the index of the inserted Node as only element
   * result[2] = Object[1] with the inserted Node as only element
   * 
   * this code was created for the insertion of an empty node at the end of the worksheet
   * @see isac.gui.treetable.CalculationModel
   */
  public Object[] addInsertionNode() {
    CalcModelNode node = new CalcModelNode(null);
    root_.add(node);
    return new Object[] {new Object[] {root_},
                         new int[] {root_.getChildCount()},
                         new Object[] {node} };
  }
}
